@@ -29,6 +29,10 @@
 
 #include "logger.h"
 #include "conversions.h"
+#include "inputstream.h"
+#include "outputstream.h"
+#include "package-version.h"
+#include "tools.h"
 
 #include <atomic>
 #include <chrono>
@@ -44,12 +48,7 @@
 #include <boost/asio/ip/address.hpp>
 #include <boost/asio/post.hpp>
 #include <boost/asio/thread_pool.hpp>
-#include <boost/iostreams/device/file_descriptor.hpp>
-#include <boost/iostreams/filtering_stream.hpp>
-#include <boost/iostreams/filtering_streambuf.hpp>
-#include <boost/iostreams/filter/bzip2.hpp>
-#include <boost/iostreams/filter/gzip.hpp>
-#include <boost/iostreams/filter/lzma.hpp>
+#include <boost/filesystem.hpp>
 #include <boost/program_options.hpp>
 #include <boost/tokenizer.hpp>
 
@@ -343,8 +342,8 @@ static bool checkFormat(boost::iostreams::filtering_ostream* outputStream,
             "SourcePort "             // 09: Source port (decimal)
             "DestinationPort "        // 10: Destination port (decimal)
             "Status "                 // 11: Status (decimal)
-            "JitterType "             // 12: Jitter type (decimal, 0 for calculation as in RFC 3550, Appendix A.8).
-            "TimeSource "             // 13: Source of the timing information (hexadecimal) as: AAQQSSHH
+            "TimeSource "             // 12: Source of the timing information (hexadecimal) as: AAQQSSHH
+            "JitterType "             // 13: Jitter type (decimal, 0 for calculation as in RFC 3550, Appendix A.8).
 
             "Packets.AppSend "        // 14: Number of packets for application send delay jitter/mean RTT computation
             "MeanDelay.AppSend "      // 15: Mean application send delay
@@ -457,6 +456,7 @@ static bool checkFormat(boost::iostreams::filtering_ostream* outputStream,
 
 // ###### Dump results file #################################################
 static bool dumpResultsFile(std::atomic<unsigned int>*                             errorCounter,
+                            unsigned long long*                                    lastTimeStamp,
                             std::set<OutputEntry*, pointer_lessthan<OutputEntry>>* outputSet,
                             boost::iostreams::filtering_ostream*                   outputStream,
                             std::mutex*                                            outputMutex,
@@ -467,52 +467,21 @@ static bool dumpResultsFile(std::atomic<unsigned int>*
                             bool*                                                  foundFormat,
                             const bool                                             checkOnly = false)
 {
-   // ====== Open input file ================================================
-#ifdef POSIX_FADV_SEQUENTIAL
-   // With fadvise() to optimise caching:
-   int handle = open(fileName.c_str(), 0, O_RDONLY);
-   if(handle < 0) {
-      HPCT_LOG(error) << "Failed to read input file " << fileName;
+   // ====== Open input stream ==============================================
+   InputStream inputStream;
+   try {
+      if(fileName != std::filesystem::path()) {
+         inputStream.openStream(fileName);
+      }
+      else {
+         inputStream.openStream(std::cin);
+      }
+   }
+   catch(std::exception& e) {
+      HPCT_LOG(fatal) << "Failed to open input file " << fileName << ": " << e.what();
       (*errorCounter)++;
       return false;
    }
-   if(posix_fadvise(handle, 0, 0, POSIX_FADV_SEQUENTIAL|POSIX_FADV_WILLNEED|POSIX_FADV_NOREUSE) < 0) {
-      HPCT_LOG(warning) << "posix_fadvise() failed:" << strerror(errno);
-   }
-
-   boost::iostreams::stream_buffer<boost::iostreams::file_descriptor_source> fstreambuffer(
-      handle,
-      boost::iostreams::file_descriptor_flags::close_handle);
-   std::istream inputFile(&fstreambuffer);
-#else
-#warning Without fadvise()
-      // Without fadvise():
-   std::ifstream                       inputFile;
-   inputFile.open(fileName, std::ios_base::in | std::ios_base::binary);
-   if(!inputFile.is_open()) {
-      HPCT_LOG(error) << "Failed to read input file " << fileName;
-      (*errorCounter)++;
-      return false;
-   }
-#endif
-
-      // ====== Prepare input stream ========================================
-   boost::iostreams::filtering_istream inputStream;
-   std::string                         extension(fileName.extension());
-   boost::algorithm::to_lower(extension);
-   if(extension == ".xz") {
-      const boost::iostreams::lzma_params params(
-         boost::iostreams::lzma::default_compression,
-         std::thread::hardware_concurrency());
-      inputStream.push(boost::iostreams::lzma_decompressor(params));
-   }
-   else if(extension == ".bz2") {
-      inputStream.push(boost::iostreams::bzip2_decompressor());
-   }
-   else if(extension == ".gz") {
-      inputStream.push(boost::iostreams::gzip_decompressor());
-   }
-   inputStream.push(inputFile);
 
    // ====== Process lines of the input file ================================
    unsigned int       version    = 0;
@@ -627,6 +596,7 @@ static bool dumpResultsFile(std::atomic<unsigned int>*
                }
 
                const std::lock_guard<std::mutex> lock(*outputMutex);
+               *lastTimeStamp = std::max(*lastTimeStamp, newEntry->TimeStamp);
                if(outputSet) {
                   auto success = outputSet->insert(newEntry);
                   if(!success.second) {
@@ -693,6 +663,7 @@ static bool dumpResultsFile(std::atomic<unsigned int>*
             }
 
             const std::lock_guard<std::mutex> lock(*outputMutex);
+            *lastTimeStamp = std::max(*lastTimeStamp, newSubEntry->TimeStamp);
             if(outputSet) {
                auto success = outputSet->insert(newSubEntry);
                if(!success.second) {
@@ -747,6 +718,8 @@ int main(int argc, char** argv)
    commandLineOptions.add_options()
       ( "help,h",
            "Print help message" )
+      ( "version",
+           "Show program version" )
 
       ( "loglevel,L",
            boost::program_options::value<unsigned int>(&logLevel)->default_value(boost::log::trivial::severity_level::info),
@@ -823,6 +796,11 @@ int main(int argc, char** argv)
                  << commandLineOptions;
        return 1;
    }
+   else if(vm.count("version")) {
+      std::cerr << "HPCT Results" << " " << HPCT_VERSION << "\n";
+      return 0;
+   }
+
    if( (separator != ' ')  &&
        (separator != '\t') &&
        (separator != ',')  &&
@@ -875,53 +853,39 @@ int main(int argc, char** argv)
    initialiseLogger(logLevel, logColor,
                     (!logFile.empty()) ? logFile.string().c_str() : nullptr);
 
-   // ====== Open output file ===============================================
-   std::string                         extension(outputFileName.extension());
-   std::ofstream                       outputFile;
-   boost::iostreams::filtering_ostream outputStream;
-   const std::filesystem::path         tmpOutputFileName = outputFileName.string() + ".tmp";
-   if(!outputFileName.empty()) {
-      outputFile.open(tmpOutputFileName, std::ios_base::out | std::ios_base::binary);
-      if(!outputFile.is_open()) {
-         HPCT_LOG(fatal) << "Failed to create output file " << tmpOutputFileName;
-         exit(1);
+   // ====== Open output stream =============================================
+   OutputStream outputStream;
+   try {
+      if(!outputFileName.empty()) {
+         outputStream.openStream(outputFileName);
       }
-      boost::algorithm::to_lower(extension);
-      if(extension == ".xz") {
-         const boost::iostreams::lzma_params params(
-            boost::iostreams::lzma::default_compression,
-            std::thread::hardware_concurrency());
-         outputStream.push(boost::iostreams::lzma_compressor(params));
+      else {
+         outputStream.openStream(std::cout);
       }
-      else if(extension == ".bz2") {
-         outputStream.push(boost::iostreams::bzip2_compressor());
-      }
-      else if(extension == ".gz") {
-         outputStream.push(boost::iostreams::gzip_compressor());
-      }
-      outputStream.push(outputFile);
    }
-   else {
-      outputStream.push(std::cout);
+   catch(std::exception& e) {
+      HPCT_LOG(fatal) << "Failed to create output file " << outputFileName << ": " << e.what();
+      exit(1);
    }
 
    // ====== Read the input files ===========================================
    const std::set<std::filesystem::path>                 inputFileNameSet(inputFileNameList.begin(), inputFileNameList.end());
    std::set<OutputEntry*, pointer_lessthan<OutputEntry>> outputSet;
    std::mutex                                            outputMutex;
+   unsigned long long                                    lastTimeStamp = 0;
+   unsigned int                                          columns       = 0;
    InputFormat                                           format { InputType::IT_Unknown };
-   unsigned int                                          columns = 0;
 
    // ------ Identify format of the first file ------------------------------
-   std::atomic<unsigned int> errorCounter = 0;
+   std::atomic<unsigned int>   errorCounter  = 0;
    const std::filesystem::path firstInputFileName = *(inputFileNameSet.begin());
    HPCT_LOG(info) << "Identifying format from " << firstInputFileName << " ...";
    bool foundFormat = false;
-   if(!dumpResultsFile(&errorCounter,
+   if(!dumpResultsFile(&errorCounter, &lastTimeStamp,
                        (sorted == true) ? &outputSet : nullptr, &outputStream, &outputMutex,
                        firstInputFileName, format, columns, separator, &foundFormat,
                        inputResultsFromStdin ? false : true)) {
-      return 1;
+      exit(1);
    }
    HPCT_LOG(info) << "Format: Type=" << (char)format.Type;
 
@@ -931,52 +895,61 @@ int main(int argc, char** argv)
                   << maxThreads << " threads ...";
    const std::chrono::system_clock::time_point t1 = std::chrono::system_clock::now();
    for(const std::filesystem::path& inputFileName : inputFileNameSet) {
-      boost::asio::post(threadPool, std::bind(dumpResultsFile,
-                        &errorCounter,
-                        (sorted == true) ? &outputSet : nullptr, &outputStream, &outputMutex,
-                        inputFileName, format, columns, separator, &foundFormat,
-                        false));
+      boost::asio::post(threadPool,
+                        std::bind(dumpResultsFile,
+                                  &errorCounter, &lastTimeStamp,
+                                  (sorted == true) ? &outputSet : nullptr, &outputStream, &outputMutex,
+                                  inputFileName, format, columns, separator, &foundFormat,
+                                  false));
    }
    threadPool.join();
    if(errorCounter > 0) {
       exit(1);
    }
+
+   // ====== Print statistics ===============================================
    const std::chrono::system_clock::time_point t2 = std::chrono::system_clock::now();
    if(sorted == true) {
       HPCT_LOG(info) << "Read " << outputSet.size() << " results rows in "
                      << std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count() << " ms";
    }
-   else {
-      HPCT_LOG(info) << "Reading input and writing output took "
-                     << std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count() << " ms";
-   }
 
    // ====== Print the results ==============================================
    if(sorted == true) {
       HPCT_LOG(info) << "Writing " << outputSet.size() << " results rows ...";
-      const std::chrono::system_clock::time_point t3 = std::chrono::system_clock::now();
       for(std::set<OutputEntry*, pointer_lessthan<OutputEntry>>::const_iterator iterator = outputSet.begin();
          iterator != outputSet.end(); iterator++) {
          outputStream << (*iterator)->Line << "\n";
          delete *iterator;
       }
-      outputStream.reset();
-
-      const std::chrono::system_clock::time_point t4 = std::chrono::system_clock::now();
-      HPCT_LOG(info) << "Wrote " << outputSet.size() << " results rows in "
-                     << std::chrono::duration_cast<std::chrono::milliseconds>(t4 - t3).count() << " ms";
    }
 
-   if(!outputFileName.empty()) {
-      try {
-         std::filesystem::rename(tmpOutputFileName, outputFileName);
-      }
-      catch(const std::exception& e) {
-         HPCT_LOG(fatal) << "Unable to rename " << tmpOutputFileName << " to " << outputFileName
-                         << ": " << e.what();
-         exit(1);
+   // ====== Close output file ==============================================
+   try {
+      outputStream.closeStream();
+      if(!outputFileName.empty()) {
+         // Set timestamp to the latest timestamp in the data.
+         // Note: the timestamp is UTC!
+         set_last_write_time(outputFileName, lastTimeStamp);
       }
    }
+   catch(const std::exception& e) {
+      HPCT_LOG(fatal) << "Failed to close output file "
+                      << outputFileName << ": " << e.what();
+      exit(1);
+   }
+
+   // ====== Print statistics ===============================================
+   const std::chrono::system_clock::time_point t3 = std::chrono::system_clock::now();
+   if(sorted == true) {
+      HPCT_LOG(info) << "Wrote " << outputSet.size() << " results rows in "
+                     << std::chrono::duration_cast<std::chrono::milliseconds>(t3 - t2).count() << " ms";
+
+   }
+   else {
+      HPCT_LOG(info) << "Reading input and writing output took "
+                     << std::chrono::duration_cast<std::chrono::milliseconds>(t3 - t1).count() << " ms";
+   }
 
    return 0;
 }
